package cn.iocoder.yudao.module.ecw.service.currencyRate;

import cn.iocoder.yudao.framework.common.exception.ServiceException;
import cn.iocoder.yudao.framework.common.pojo.PageParam;
import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.module.ecw.dal.dataobject.currencyRate.CurrencyRateDO;
import cn.iocoder.yudao.module.ecw.dal.dataobject.currencyRate.CurrencyRateLogDO;
import cn.iocoder.yudao.module.ecw.vo.currencyRate.CurrencyRateCreateReqVO;
import cn.iocoder.yudao.module.ecw.vo.currencyRate.CurrencyRateQueryParamVO;
import cn.iocoder.yudao.module.ecw.vo.currencyRate.CurrencyRateUpdateReqVO;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.time.LocalDate;
import java.util.List;
import java.util.Optional;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_AMOUNT_INVALID;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_EXPIRED;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_NOT_EXISTS;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_TOO_SMALL;

/**
 * 货币汇率服务
 */
public interface CurrencyRateService {
    /**
     * 汇率精度。
     */
    int RATE_SCALE = 8;

    /**
     * 汇率四舍五入。
     */
    RoundingMode RATE_ROUNDING = RoundingMode.HALF_UP;

    /**
     * 计算两个金额的汇率。
     *
     * @param source 原金额
     * @param target 目标金额
     * @return 汇率
     * @throws ServiceException 金额小于或等零，或者汇率小于精度下最小值
     */
    default BigDecimal rate(BigDecimal source, BigDecimal target) throws ServiceException {
        if (source.compareTo(BigDecimal.ZERO) <= 0 || target.compareTo(BigDecimal.ZERO) <= 0) {
            throw exception(CURRENCY_RATE_AMOUNT_INVALID);
        }

        return Optional.of(target.divide(source, RATE_SCALE, RATE_ROUNDING))
                .filter(it -> it.compareTo(BigDecimal.ZERO) > 0)
                .orElseThrow(() -> exception(CURRENCY_RATE_TOO_SMALL));
    }


    /**
     * 获得两个币种的兑换基础值。
     *
     * @param sourceId 原币种编号
     * @param targetId 支付币种编号
     * @return 获得两个币种的兑换基础值
     * @throws ServiceException 汇率不存在或过期
     */
    default CurrencyRateDO exchangeBaseValue(long sourceId, long targetId) throws ServiceException {
        CurrencyRateDO entity = Optional.ofNullable(find(sourceId, targetId))
                .orElseThrow(() -> exception(CURRENCY_RATE_NOT_EXISTS));
        LocalDate now = LocalDate.now();

        if (now.isAfter(entity.getExpiration())) {
            throw exception(CURRENCY_RATE_EXPIRED, String.valueOf(sourceId), String.valueOf(targetId));
        }

        return entity;
    }

    /**
     * 计算两个币种的汇率。
     *
     * @param sourceId 原币种编号
     * @param targetId 支付币种编号
     * @return 汇率比值，为空则表示汇率不存在
     * @throws ServiceException 汇率不存在或过期
     */
    default BigDecimal rate(long sourceId, long targetId) throws ServiceException {
        CurrencyRateDO entity = Optional.ofNullable(find(sourceId, targetId))
                .orElseThrow(() -> exception(CURRENCY_RATE_NOT_EXISTS));
        LocalDate now = LocalDate.now();

        if (now.isAfter(entity.getExpiration())) {
            throw exception(CURRENCY_RATE_EXPIRED, String.valueOf(sourceId), String.valueOf(targetId));
        }

        return rate(entity.getSourceAmount(), entity.getTargetAmount());
    }

    /**
     * 计算两个币种的汇率。
     *
     * @param sourceId 原币种编号
     * @param targetId 支付币种编号
     * @param tryOpposite 如果不存在则使用反向汇率计算
     * @return 汇率比值，为空则表示汇率不存在
     * @throws ServiceException 汇率不存在或过期或反向汇率过小
     */
    default BigDecimal rate(long sourceId, long targetId, boolean tryOpposite) throws ServiceException {
        try {
            return rate(sourceId, targetId);
        } catch (ServiceException e) {
            if (!tryOpposite) {
                throw e;
            }

            return BigDecimal.ONE.divide(rate(targetId, sourceId), RATE_SCALE, RATE_ROUNDING);
        }
    }

    /**
     * 创建新汇率。
     *
     * @param req 创建请求
     * @throws ServiceException 汇率已存在或金额国家等参数不合法
     */
    void create(CurrencyRateCreateReqVO req);

    /**
     * 获取汇率信息。
     *
     * @param sourceId 原币种编号
     * @param targetId 支付币种编号
     * @return 汇率信息，可能为空
     */
    CurrencyRateDO find(long sourceId, long targetId);

    /**
     * 分页查询汇率。
     *
     * @param param 分页参数
     * @return 分页结果
     */
    PageResult<CurrencyRateDO> query(CurrencyRateQueryParamVO param);


    /**
     * 查询全部汇率数据
     *
     * @return 结果
     */
    List<CurrencyRateDO> queryAll();

    /**
     * 更新汇率信息。
     *
     * @param id 汇率信息编号
     * @param req 更新请求
     * @throws ServiceException 汇率不存在或金额等参数不合法
     */
    void update(long id, CurrencyRateUpdateReqVO req);

    /**
     * 删除汇率信息。
     *
     * @param id 汇率信息编号
     */
    void delete(long id);

    /**
     * 分页查询汇率变更历史。
     *
     * @param id 汇率信息编号
     * @param param 分页参数
     * @return 分页结果
     */
    PageResult<CurrencyRateLogDO> history(long id, PageParam param);
}
