package cn.iocoder.yudao.module.ecw.service.currencyRate;

import cn.iocoder.yudao.framework.common.pojo.PageParam;
import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.module.ecw.convert.currencyRate.CurrencyRateConvert;
import cn.iocoder.yudao.module.ecw.dal.dataobject.currencyRate.CurrencyRateDO;
import cn.iocoder.yudao.module.ecw.dal.dataobject.currencyRate.CurrencyRateLogDO;
import cn.iocoder.yudao.module.ecw.dal.mysql.currencyRate.CurrencyRateLogMapper;
import cn.iocoder.yudao.module.ecw.dal.mysql.currencyRate.CurrencyRateMapper;
import cn.iocoder.yudao.module.ecw.service.country.CountryService;
import cn.iocoder.yudao.module.ecw.service.currency.CurrencyService;
import cn.iocoder.yudao.module.ecw.vo.currencyRate.CurrencyRateCreateReqVO;
import cn.iocoder.yudao.module.ecw.vo.currencyRate.CurrencyRateQueryParamVO;
import cn.iocoder.yudao.module.ecw.vo.currencyRate.CurrencyRateUpdateReqVO;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.List;
import java.util.Optional;
import java.util.Set;

import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_NOT_EXISTS;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_COUNTRIES_INVALID;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_EXISTS;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_NOT_EXISTS;
import static cn.iocoder.yudao.module.ecw.enums.ErrorCodeConstants.CURRENCY_RATE_PAIR_INVALID;

@Service
public class CurrencyRateServiceImpl implements CurrencyRateService {
    @Autowired
    private CountryService countryService;

    @Autowired
    private CurrencyService currencyService;

    @Autowired
    private CurrencyRateMapper mapper;

    @Autowired
    private CurrencyRateLogMapper logMapper;

    // 检查指定币种是否存在
    private void currency(long id) {
        Optional.ofNullable(currencyService.getCurrency((int) id))
                .orElseThrow(() -> exception(CURRENCY_NOT_EXISTS));
    }

    @Override
    public void create(CurrencyRateCreateReqVO req) {
        long sourceId = req.getSourceId();
        long targetId = req.getTargetId();
        Set<Long> countries = req.getCountries();

        if (sourceId == targetId) {
            throw exception(CURRENCY_RATE_PAIR_INVALID);
        }

        rate(req.getSourceAmount(), req.getTargetAmount()); // check rate

        if (countries.isEmpty() || (countries.size() != countryService.getCountryList(countries).size())) {
            throw exception(CURRENCY_RATE_COUNTRIES_INVALID);
        }

        // check if currencies exist
        currency(sourceId);
        currency(targetId);

        if (find(sourceId, targetId) != null) {
            throw exception(CURRENCY_RATE_EXISTS);
        }

        mapper.insert(CurrencyRateConvert.INSTANCE.from(req));
    }

    @Override
    public CurrencyRateDO find(long sourceId, long targetId) {
        LambdaQueryWrapper<CurrencyRateDO> wrapper = Wrappers.lambdaQuery(CurrencyRateDO.class)
                .eq(CurrencyRateDO::getSourceId, sourceId)
                .eq(CurrencyRateDO::getTargetId, targetId);
        return mapper.selectOne(wrapper);
    }

    @Override
    public PageResult<CurrencyRateDO> query(CurrencyRateQueryParamVO param) {
        return mapper.selectPage(param);
    }

    @Override
    public List<CurrencyRateDO> queryAll() {
        return mapper.selectList();
    }

    @Override
    public void update(long id, CurrencyRateUpdateReqVO req) {
        rate(req.getSourceAmount(), req.getTargetAmount()); // check rate

        CurrencyRateLogDO change = new CurrencyRateLogDO();
        CurrencyRateDO saved = Optional.ofNullable(mapper.selectById(id))
                .orElseThrow(() -> exception(CURRENCY_RATE_NOT_EXISTS));

        // 修改前
        change.setSourceAmount1(saved.getSourceAmount());
        change.setTargetAmount1(saved.getTargetAmount());
        change.setExpiration1(saved.getExpiration());

        // 修改后
        change.setSourceAmount2(req.getSourceAmount());
        change.setTargetAmount2(req.getTargetAmount());
        change.setExpiration2(req.getExpiration());

        change.setRateId(id);
        change.setRemarks(req.getRemarks());
        logMapper.insert(change);

        CurrencyRateConvert.INSTANCE.update(saved, req);
        mapper.updateById(saved);
    }

    @Override
    public void delete(long id) {
        mapper.deleteById(id);
    }

    @Override
    public PageResult<CurrencyRateLogDO> history(long id, PageParam param) {
        LambdaQueryWrapper<CurrencyRateLogDO> wrapper = Wrappers.lambdaQuery(CurrencyRateLogDO.class)
                .eq(CurrencyRateLogDO::getRateId, id)
                .orderByDesc(CurrencyRateLogDO::getCreateTime);
        return logMapper.selectPage(param, wrapper);
    }
}
